@@ -3,7 +3,7 @@
 #' Get, list, create, or delete blob containers.
 #'
 #' @param endpoint Either a blob endpoint object as created by [storage_endpoint], or a character string giving the URL of the endpoint.
-#' @param key,sas If an endpoint object is not supplied, authentication details. If a key is provided, the SAS is not used. If neither an access key nor a SAS are provided, only public (anonymous) access to the share is possible.
+#' @param key,token,sas If an endpoint object is not supplied, authentication credentials: either an access key, an Azure Active Directory (AAD) token, or a SAS, in that order of priority. If no authentication credentials are provided, only public (anonymous) access to the share is possible.
 #' @param api_version If an endpoint object is not supplied, the storage API version to use when interacting with the host. Currently defaults to `"2018-03-28"`.
 #' @param name The name of the blob container to get, create, or delete.
 #' @param confirm For deleting a container, whether to ask for confirmation.
@@ -15,14 +15,17 @@
 #' @details
 #' You can call these functions in a couple of ways: by passing the full URL of the share, or by passing the endpoint object and the name of the container as a string.
 #'
-#' Currently (as of December 2018), if the storage account has hierarchical namespaces enabled, the blob API for the account is disabled. The blob endpoint is still accessible, but blob operations on the endpoint will fail. Full interoperability between blobs and ADLS is planned for 2019.
+#' If authenticating via AAD, you can supply the token either as a string, or as an object of class AzureToken, created via [AzureRMR::get_azure_token]. The latter is the recommended way of doing it, as it allows for automatic refreshing of expired tokens.
+#'
+#' Currently (as of February 2019), if hierarchical namespaces are enabled on a storage account, the blob API for the account is disabled. The blob endpoint is still accessible, but blob operations on the endpoint will fail. Full interoperability between blobs and ADLSgen2 is planned for later in 2019.
 #'
 #' @return
 #' For `blob_container` and `create_blob_container`, an S3 object representing an existing or created container respectively.
 #'
 #' For `list_blob_containers`, a list of such objects.
 #'
-#' @seealso [storage_endpoint], [az_storage]
+#' @seealso
+#' [storage_endpoint], [az_storage], [storage_container]
 #'
 #' @examples
 #' \dontrun{
@@ -42,6 +45,13 @@
 #' create_blob_container("https://mystorage.blob.core.windows.net/newcontainer", key="access_key")
 #' delete_blob_container("https://mystorage.blob.core.windows.net/newcontainer", key="access_key")
 #'
+#' # authenticating via AAD
+#' token <- AzureRMR::get_azure_token(resource="https://storage.azure.com/",
+#'     tenant="myaadtenant",
+#'     app="myappid",
+#'     password="mypassword")
+#' blob_container("https://mystorage.blob.core.windows.net/mycontainer", token=token)
+#'
 #' }
 #' @rdname blob_container
 #' @export
@@ -52,11 +62,11 @@ blob_container <- function(endpoint, ...)
 
 #' @rdname blob_container
 #' @export
-blob_container.character <- function(endpoint, key=NULL, sas=NULL,
+blob_container.character <- function(endpoint, key=NULL, token=NULL, sas=NULL,
                                      api_version=getOption("azure_storage_api_version"),
                                      ...)
 {
-    do.call(blob_container, generate_endpoint_container(endpoint, key, sas, api_version))
+    do.call(blob_container, generate_endpoint_container(endpoint, key, token, sas, api_version))
 }
 
 #' @rdname blob_container
@@ -64,7 +74,7 @@ blob_container.character <- function(endpoint, key=NULL, sas=NULL,
 blob_container.blob_endpoint <- function(endpoint, name, ...)
 {
     obj <- list(name=name, endpoint=endpoint)
-    class(obj) <- "blob_container"
+    class(obj) <- c("blob_container", "storage_container")
     obj
 }
 
@@ -74,12 +84,22 @@ print.blob_container <- function(x, ...)
 {
     cat("Azure blob container '", x$name, "'\n", sep="")
     cat(sprintf("URL: %s\n", paste0(x$endpoint$url, x$name)))
+
     if(!is_empty(x$endpoint$key))
         cat("Access key: <hidden>\n")
     else cat("Access key: <none supplied>\n")
+
+    if(!is_empty(x$endpoint$token))
+    {
+        cat("Azure Active Directory access token:\n")
+        print(x$endpoint$token)
+    }
+    else cat("Azure Active Directory access token: <none supplied>\n")
+
     if(!is_empty(x$endpoint$sas))
         cat("Account shared access signature: <hidden>\n")
     else cat("Account shared access signature: <none supplied>\n")
+
     cat(sprintf("Storage API version: %s\n", x$endpoint$api_version))
     invisible(x)
 }
@@ -94,11 +114,11 @@ list_blob_containers <- function(endpoint, ...)
 
 #' @rdname blob_container
 #' @export
-list_blob_containers.character <- function(endpoint, key=NULL, sas=NULL,
+list_blob_containers.character <- function(endpoint, key=NULL, token=NULL, sas=NULL,
                                            api_version=getOption("azure_storage_api_version"),
                                            ...)
 {
-    do.call(list_blob_containers, generate_endpoint_container(endpoint, key, sas, api_version))
+    do.call(list_blob_containers, generate_endpoint_container(endpoint, key, token, sas, api_version))
 }
 
 #' @rdname blob_container
@@ -106,7 +126,8 @@ list_blob_containers.character <- function(endpoint, key=NULL, sas=NULL,
 list_blob_containers.blob_endpoint <- function(endpoint, ...)
 {
     lst <- do_storage_call(endpoint$url, "/", options=list(comp="list"),
-                           key=endpoint$key, sas=endpoint$sas, api_version=endpoint$api_version)
+                           key=endpoint$key, token=endpoint$token, sas=endpoint$sas,
+                           api_version=endpoint$api_version)
 
     lst <- lapply(lst$Containers, function(cont) blob_container(endpoint, cont$Name[[1]]))
     named_list(lst)
@@ -123,11 +144,11 @@ create_blob_container <- function(endpoint, ...)
 
 #' @rdname blob_container
 #' @export
-create_blob_container.character <- function(endpoint, key=NULL, sas=NULL,
+create_blob_container.character <- function(endpoint, key=NULL, token=NULL, sas=NULL,
                                             api_version=getOption("azure_storage_api_version"),
                                             ...)
 {
-    endp <- generate_endpoint_container(endpoint, key, sas, api_version)
+    endp <- generate_endpoint_container(endpoint, key, token, sas, api_version)
     create_blob_container(endp$endpoint, endp$name, ...)
 }
 
@@ -163,11 +184,11 @@ delete_blob_container <- function(endpoint, ...)
 
 #' @rdname blob_container
 #' @export
-delete_blob_container.character <- function(endpoint, key=NULL, sas=NULL,
+delete_blob_container.character <- function(endpoint, key=NULL, token=NULL, sas=NULL,
                                             api_version=getOption("azure_storage_api_version"),
                                             ...)
 {
-    endp <- generate_endpoint_container(endpoint, key, sas, api_version)
+    endp <- generate_endpoint_container(endpoint, key, token, sas, api_version)
     delete_blob_container(endp$endpoint, endp$name, ...)
 }
 
@@ -198,26 +219,37 @@ delete_blob_container.blob_endpoint <- function(endpoint, name, confirm=TRUE, le
 }
 
 
-#' Operations on a blob container
+#' Operations on a blob container or blob
 #'
 #' Upload, download, or delete a blob; list blobs in a container.
 #'
 #' @param container A blob container object.
 #' @param blob A string naming a blob.
-#' @param src,dest The source and destination filenames for uploading and downloading. Paths are allowed.
+#' @param src,dest The source and destination files for uploading and downloading. See 'Details' below.For uploading, `src` can also be a [textConnection] or [rawConnection] object to allow transferring in-memory R objects without creating a temporary file. For downloading, 
 #' @param info For `list_blobs`, level of detail about each blob to return: a vector of names only; the name, size and last-modified date (default); or all information.
 #' @param confirm Whether to ask for confirmation on deleting a blob.
 #' @param blocksize The number of bytes to upload per HTTP(S) request.
 #' @param lease The lease for a blob, if present.
 #' @param type When uploading, the type of blob to create. Currently only block blobs are supported.
 #' @param overwrite When downloading, whether to overwrite an existing destination file.
+#' @param use_azcopy Whether to use the AzCopy utility from Microsoft to do the transfer, rather than doing it in R.
+#' @param max_concurrent_transfers For `multiupload_blob` and `multidownload_blob`, the maximum number of concurrent file transfers. Each concurrent file transfer requires a separate R process, so limit this if you are low on memory.
 #' @param prefix For `list_blobs`, filters the result to return only blobs whose name begins with this prefix.
 #'
+#' @details
+#' `upload_blob` and `download_blob` are the workhorse file transfer functions for blobs. They each take as inputs a _single_ filename or connection as the source for uploading/downloading, and a single filename as the destination.
+#'
+#' `multiupload_blob` and `multidownload_blob` are functions for uploading and downloading _multiple_ blobs at once. They parallelise file transfers by deploying a pool of R processes in the background, which can lead to significantly greater efficiency when transferring many small files. They take as input a wildcard pattern as the source, which expands to one or more files. The `dest` argument should be a directory.
+#'
+#' The file transfer functions also support working with connections to allow transferring R objects without creating temporary files. For uploading, `src` can be a [textConnection] or [rawConnection] object. For downloading, `dest` can be NULL or a `rawConnection` object. In the former case, the downloaded data is returned as a raw vector, and for the latter, it will be placed into the connection. See the examples below.
+#'
 #' @return
-#' For `list_blobs`, details on the blobs in the container.
+#' For `list_blobs`, details on the blobs in the container. For `download_blob`, if `dest=NULL`, the contents of the downloaded blob as a raw vector.
 #'
 #' @seealso
-#' [blob_container], [az_storage]
+#' [blob_container], [az_storage], [storage_download], [call_azcopy]
+#'
+#' [AzCopy version 10 on GitHub](https://github.com/Azure/azure-storage-azcopy)
 #'
 #' @examples
 #' \dontrun{
@@ -231,6 +263,28 @@ delete_blob_container.blob_endpoint <- function(endpoint, name, confirm=TRUE, le
 #'
 #' delete_blob(cont, "bigfile.zip")
 #'
+#' # uploading/downloading multiple files at once
+#' multiupload_blob(cont, "/data/logfiles/*.zip", "/uploaded_data")
+#' multiupload_blob(cont, "myproj/*")  # no dest directory uploads to root
+#' multidownload_blob(cont, "jan*.*", "/data/january")
+#'
+#' # uploading serialized R objects via connections
+#' json <- jsonlite::toJSON(iris, pretty=TRUE, auto_unbox=TRUE)
+#' con <- textConnection(json)
+#' upload_blob(cont, con, "iris.json")
+#'
+#' rds <- serialize(iris, NULL)
+#' con <- rawConnection(rds)
+#' upload_blob(cont, con, "iris.rds")
+#'
+#' # downloading files into memory: as a raw vector, and via a connection
+#' rawvec <- download_blob(cont, "iris.json", NULL)
+#' rawToChar(rawvec)
+#'
+#' con <- rawConnection(raw(0), "r+")
+#' download_blob(cont, "iris.rds", con)
+#' unserialize(con)
+#'
 #' }
 #' @rdname blob
 #' @export
@@ -277,61 +331,45 @@ list_blobs <- function(container, info=c("partial", "name", "all"),
 
 #' @rdname blob
 #' @export
-upload_blob <- function(container, src, dest, type="BlockBlob", blocksize=2^24, lease=NULL)
+upload_blob <- function(container, src, dest, type="BlockBlob", blocksize=2^24, lease=NULL,
+                        use_azcopy=FALSE)
 {
-    if(type != "BlockBlob")
-        stop("Only block blobs currently supported")
-    content_type <- mime::guess_type(src)
-    headers <- list("x-ms-blob-type"=type)
-    if(!is.null(lease))
-        headers[["x-ms-lease-id"]] <- as.character(lease)
-
-    con <- if(inherits(src, "textConnection"))
-        rawConnection(charToRaw(paste0(readLines(src), collapse="\n")))
-    else file(src, open="rb")
-    on.exit(close(con))
-
-    # upload each block
-    blocklist <- list()
-    i <- 1
-    while(1)
-    {
-        body <- readBin(con, "raw", blocksize)
-        thisblock <- length(body)
-        if(thisblock == 0)
-            break
-
-        # ensure content-length is never exponential notation
-        headers[["content-length"]] <- sprintf("%.0f", thisblock)
-        id <- openssl::base64_encode(sprintf("%s-%010d", dest, i))
-        opts <- list(comp="block", blockid=id)
-
-        do_container_op(container, dest, headers=headers, body=body, options=opts, http_verb="PUT")
-
-        blocklist <- c(blocklist, list(Latest=list(id)))
-        i <- i + 1
-    }
-
-    # update block list
-    body <- as.character(xml2::as_xml_document(list(BlockList=blocklist)))
-    headers <- list("content-length"=nchar(body))
-    do_container_op(container, dest, headers=headers, body=body, options=list(comp="blocklist"),
-                    http_verb="PUT")
-
-    # set content type
-    do_container_op(container, dest, headers=list("x-ms-blob-content-type"=content_type),
-                    options=list(comp="properties"),
-                    http_verb="PUT")
+    if(use_azcopy)
+        azcopy_upload(container, src, dest, type=type, blocksize=blocksize, lease=lease)
+    else upload_blob_internal(container, src, dest, type=type, blocksize=blocksize, lease=lease)
 }
 
 #' @rdname blob
 #' @export
-download_blob <- function(container, src, dest, overwrite=FALSE, lease=NULL)
+multiupload_blob <- function(container, src, dest, type="BlockBlob", blocksize=2^24, lease=NULL,
+                             use_azcopy=FALSE,
+                             max_concurrent_transfers=10)
 {
-    headers <- list()
-    if(!is.null(lease))
-        headers[["x-ms-lease-id"]] <- as.character(lease)
-    do_container_op(container, src, headers=headers, config=httr::write_disk(dest, overwrite))
+    if(use_azcopy)
+        azcopy_upload(container, src, dest, type=type, blocksize=blocksize, lease=lease)
+    else multiupload_blob_internal(container, src, dest, type=type, blocksize=blocksize, lease=lease,
+                                   max_concurrent_transfers=max_concurrent_transfers)
+}
+
+#' @rdname blob
+#' @export
+download_blob <- function(container, src, dest, overwrite=FALSE, lease=NULL, use_azcopy=FALSE)
+{
+    if(use_azcopy)
+        azcopy_download(container, src, dest, overwrite=overwrite, lease=lease)
+    else download_blob_internal(container, src, dest, overwrite=overwrite, lease=lease)
+}
+
+#' @rdname blob
+#' @export
+multidownload_blob <- function(container, src, dest, overwrite=FALSE, lease=NULL,
+                               use_azcopy=FALSE,
+                               max_concurrent_transfers=10)
+{
+    if(use_azcopy)
+        azcopy_download(container, src, dest, overwrite=overwrite, lease=lease)
+    else multidownload_blob_internal(container, src, dest, overwrite=overwrite, lease=lease,
+                                     max_concurrent_transfers=max_concurrent_transfers)
 }
 
 #' @rdname blob
